#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of LIBSGM
#
#     https://github.com/CNES/Pandora_libsgm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
import libsgm_python.sgm_python as sgm
import numpy as np


class TestSgmPython(unittest.TestCase):

    """"
    Test Python version of LibSGM
    """

    ###############################################################
    # Sgm
    ###############################################################

    def test_compute_lr(self):
        p1 = 3
        p2 = 4

        cv_in_2d_front = np.array([[1, 2, 3, 4], # same d
                                   [5, 6, 7, 8], # |d' - d| = 1
                                   [9, 10, 11, 12]]) # |d' - d| > 1

        lr_2d_previous = np.array([[1, 1, 1, 0],
                                   [1, 8, 2, 9],
                                   [9, 10, 11, 1]])

        p1_in_1d = p1 * np.ones(3)
        p2_in_1d = p2 * np.ones(3)

        # disp
        d = 1

        lr_d = sgm.compute_lr(cv_in_2d_front, lr_2d_previous, d, p1_in_1d, p2_in_1d)

        real_lr_d = np.array([3, 9, 14])
        np.testing.assert_array_equal(lr_d, real_lr_d)

    def test_sgm_middle_value_invalid(self):
        p1 = 8
        p2 = 32

        cv_in = np.array([[[1,15,20], [14,16,6], [8,19,8]],
                          [[13,11,3], [np.nan,np.nan,np.nan], [16,4,12]],
                          [[18,2,17], [23,7,1], [5,20,14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[0, 1], [1, 0], [1, 1], [1, -1], [0, -1], [-1, 0], [-1, -1], [-1, 1]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=False, overcounting=False)

        # invalid value : nan
        self.assertTrue(np.isnan(cv_out["cv"][1,1,1]))

    def test_sgm_middle_value_invalid_overcounting(self):
        p1 = 8
        p2 = 32

        cv_in = np.array([[[1, 15, 20], [14, 16, 6], [8, 19, 8]],
                          [[13, 11, 3], [np.nan, np.nan, np.nan], [16, 4, 12]],
                          [[18, 2, 17], [23, 7, 1], [5, 20, 14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[0, 1], [1, 0], [1, 1], [1, -1], [0, -1], [-1, 0], [-1, -1], [-1, 1]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=False, overcounting=True)

        # invalid value : nan
        self.assertTrue(np.isnan(cv_out["cv"][1, 1, 1]))

    def test_sgm_middle_value_lr_0_1(self):
        p1 = 8
        p2 = 32
        cv_in = np.array([[[1, 15, 20], [14, 16, 6], [8, 19, 8]],
                          [[13, 11, 3], [22, 12, 9], [16, 4, 12]],
                          [[18, 2, 17], [23, 7, 1], [5, 20, 14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[0, 1]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=False, overcounting=False)

        # invalid value : nan
        self.assertEqual(cv_out["cv"][1, 1, 1], 20)

    def test_sgm_middle_value_lr_1_0(self):
        p1 = 8
        p2 = 32
        cv_in = np.array([[[1, 15, 20], [14, 16, 6], [8, 19, 8]],
                          [[13, 11, 3], [22, 12, 9], [16, 4, 12]],
                          [[18, 2, 17], [23, 7, 1], [5, 20, 14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[1, 0]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=False, overcounting=False)

        # invalid value : nan
        self.assertEqual(cv_out["cv"][1, 1, 1], 20)

    def test_sgm_middle_value_lr_minus1_0(self):
        p1 = 8
        p2 = 32
        cv_in = np.array([[[1, 15, 20], [14, 16, 6], [8, 19, 8]],
                          [[13, 11, 3], [22, 12, 9], [16, 4, 12]],
                          [[18, 2, 17], [23, 7, 1], [5, 20, 14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[-1, 0]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=False, overcounting=False)

        # invalid value : nan
        self.assertEqual(cv_out["cv"][1, 1, 1], 18)

    def test_sgm_middle_value_lr_0_minus1(self):
        p1 = 8
        p2 = 32
        cv_in = np.array([[[1, 15, 20], [14, 16, 6], [8, 19, 8]],
                          [[13, 11, 3], [22, 12, 9], [16, 4, 12]],
                          [[18, 2, 17], [23, 7, 1], [5, 20, 14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[0, -1]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=False, overcounting=False)

        # invalid value : nan
        self.assertEqual(cv_out["cv"][1, 1, 1], 12)

    def test_sgm_middle_value_lr_1_1(self):
        p1 = 8
        p2 = 32
        cv_in = np.array([[[1, 15, 20], [14, 16, 6], [8, 19, 8]],
                          [[13, 11, 3], [22, 12, 9], [16, 4, 12]],
                          [[18, 2, 17], [23, 7, 1], [5, 20, 14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[1, 1]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=False, overcounting=False)

        # invalid value : nan
        self.assertEqual(cv_out["cv"][1, 1, 1], 20)

    def test_sgm_middle_value_lr_minus1_1(self):
        p1 = 8
        p2 = 32
        cv_in = np.array([[[1, 15, 20], [14, 16, 6], [8, 19, 8]],
                          [[13, 11, 3], [22, 12, 9], [16, 4, 12]],
                          [[18, 2, 17], [23, 7, 1], [5, 20, 14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[-1, 1]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=False, overcounting=False)

        # invalid value : nan
        self.assertEqual(cv_out["cv"][1, 1, 1], 12)

    def test_sgm_middle_value_lr_1_minus1(self):
        p1 = 8
        p2 = 32
        cv_in = np.array([[[1, 15, 20], [14, 16, 6], [8, 19, 8]],
                          [[13, 11, 3], [22, 12, 9], [16, 4, 12]],
                          [[18, 2, 17], [23, 7, 1], [5, 20, 14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[1, -1]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=False, overcounting=False)

        # invalid value : nan
        self.assertEqual(cv_out["cv"][1, 1, 1], 20)

    def test_sgm_middle_value_lr_minus1_minus1(self):
        p1 = 8
        p2 = 32
        cv_in = np.array([[[1, 15, 20], [14, 16, 6], [8, 19, 8]],
                          [[13, 11, 3], [22, 12, 9], [16, 4, 12]],
                          [[18, 2, 17], [23, 7, 1], [5, 20, 14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[-1, -1]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=False, overcounting=False)

        # invalid value : nan
        self.assertEqual(cv_out["cv"][1, 1, 1], 20)

    def test_sgm_middle_value(self):
        p1 = 8
        p2 = 32
        cv_in = np.array([[[1, 15, 20], [14, 16, 6], [8, 19, 8]],
                          [[13, 11, 3], [22, 12, 9], [16, 4, 12]],
                          [[18, 2, 17], [23, 7, 1], [5, 20, 14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[0, 1], [1, 0], [1, 1], [1, -1], [0, -1], [-1, 0], [-1, -1], [-1, 1]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=False, overcounting=False)

        # invalid value : nan
        self.assertEqual(cv_out["cv"][1, 1, 1], 142)

    def test_sgm_middle_value_overcounting(self):
        p1 = 8
        p2 = 32

        cv_in = np.array([[[1, 15, 20], [14, 16, 6], [8, 19, 8]],
                          [[13, 11, 3], [22, 12, 9], [16, 4, 12]],
                          [[18, 2, 17], [23, 7, 1], [5, 20, 14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[0, 1], [1, 0], [1, 1], [1, -1], [0, -1], [-1, 0], [-1, -1], [-1, 1]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=False, overcounting=True)

        # invalid value : nan
        self.assertEqual(cv_out["cv"][1, 1, 1], 58)

    def test_sgm_middle_value_min_cost(self):
        p1 = 8
        p2 = 32

        cv_in = np.array([[[1, 15, 20], [14, 16, 6], [8, 19, 8]],
                          [[13, 11, 3], [22, 12, 9], [16, 4, 12]],
                          [[18, 2, 17], [23, 7, 1], [5, 20, 14]]])

        p1_in = p1 * np.ones((3, 3, 8))
        p2_in = p2 * np.ones((3, 3, 8))

        directions = [[0, 1], [1, 0], [1, 1], [1, -1], [0, -1], [-1, 0], [-1, -1], [-1, 1]]

        cv_out = sgm.run_sgm(cv_in, p1_in, p2_in, directions, cost_paths=True, overcounting=False)

        # invalid value : nan
        self.assertEqual(cv_out["cv"][1, 1, 1], 142)
        self.assertEqual(cv_out["cv_min"][1, 1, 1], 2)


#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of LIBSGM
#
#     https://github.com/CNES/Pandora_libsgm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions to execute python SGM library.
"""

import numpy as np
import copy
from typing import List, Tuple


class LrManager:
    """
    This class manages parallel planes indices moving along given direction
    """
    # initialize
    def __init__(self, cv_shape: List, direction: Tuple):
        """
        Initialise LrManager

        :param cv_shape: shape of cost volume
        :type cv_shape:  List
        :param direction: direction used (d_i, d_j)
        :type direction:  Tuple
        """
        self.cv_shape = cv_shape
        self.d = direction

        # init working matrices
        self.planes_front = []
        self.planes_previous = []

        # store partial lr
        self.current_lr = None
        self.previous_lr = None

        # start and end indices to manage duplicated corners when two planes are used
        start = 0
        end = cv_shape[0]
        if direction[0] > 0:
            # TODO deal with directions > 1 : several planes, be careful about duplicated points
            self.planes_front.append({"i": np.array([0]), "j": np.array(list(range(0, cv_shape[1])))})
            start = 1
        if direction[0] < 0:
            self.planes_front.append({"i": np.array([cv_shape[0] - 1]), "j": np.array(list(range(0, cv_shape[1])))})
            end = end - 1
        if direction[1] > 0:
            self.planes_front.append({"i": np.array(list(range(start, end))), "j": np.array([0])})
        if direction[1] < 0:
            self.planes_front.append({"i": np.array(list(range(start, end))), "j": np.array([cv_shape[1] - 1])})

    def next(self) -> None:
        """
        Update the state of the planes, moves forward
        """
        self.planes_previous = copy.deepcopy(self.planes_front)
        self.previous_lr = self.current_lr
        delete_items_indexes = []
        for p in range(len(self.planes_front)):
            self.planes_front[p]["i"] += self.d[0]
            self.planes_front[p]["j"] += self.d[1]

            # get invalid indexes
            indexes_invalid_i = np.where(
                (self.planes_front[p]["i"] < 0) + (self.planes_front[p]["i"] >= self.cv_shape[0]))
            indexes_invalid_j = np.where(
                (self.planes_front[p]["j"] < 0) + (self.planes_front[p]["j"] >= self.cv_shape[1]))

            # remove invalid indexes
            self.planes_front[p]["i"] = np.delete(self.planes_front[p]["i"], indexes_invalid_i, axis=0)
            self.planes_previous[p]["i"] = np.delete(self.planes_previous[p]["i"], indexes_invalid_i, axis=0)

            self.planes_front[p]["j"] = np.delete(self.planes_front[p]["j"], indexes_invalid_j, axis=0)
            self.planes_previous[p]["j"] = np.delete(self.planes_previous[p]["j"], indexes_invalid_j, axis=0)

            if self.planes_front[p]["i"].shape[0] == 0 or self.planes_front[p]["j"].shape[0] == 0:
                # delete all object
                delete_items_indexes.append(p)
            else:
                # delete in stored lr, as list are of size (1) (n) or (m) (1), only one delete is effective

                if self.previous_lr is not None:
                    self.previous_lr[p] = np.delete(self.previous_lr[p], indexes_invalid_i, axis=0)
                    self.previous_lr[p] = np.delete(self.previous_lr[p], indexes_invalid_j, axis=0)

        # delete items
        for p in sorted(delete_items_indexes, reverse=True):
            self.planes_front.pop(p)
            self.planes_previous.pop(p)
            if self.previous_lr is not None:
                self.previous_lr.pop(p)

    def set_current_lr(self, lr_s: np.ndarray) -> None:
        """
        Set current lr

        :param lr_s: partial cost lr
        :type lr_s:  np.ndarray
        """
        self.current_lr = lr_s

    def get_previous_lr(self, num_plane: int) -> np.ndarray:
        """
        Get previous lr

        :param num_plane: numero of plane
        :type num_plane:  int
        :return: previous partial cost lr
        :rtype: np.ndarray
        """
        return self.previous_lr[num_plane]



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of LIBSGM
#
#     https://github.com/CNES/Pandora_libsgm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions to execute SGM library.
"""

cimport cython
from libc.stdlib cimport *
from libc.stdint cimport *
from libcpp.string cimport string
from libcpp cimport bool
from cpython cimport PyObject , Py_INCREF

import numpy as np
cimport numpy as np
cimport openmp

#Numpy initialization to avoid segfaults
np.import_array()

#Fused type to allow multiple types definitions
#The following list represent Cost volume types avalaible
#You can add another one
ctypedef fused my_cv_type:
    uint8_t
    float

# ArrayWrapper class  to deallocate our array
# when Python Object is deleted

cdef class Array3DWrapper:

    cdef void* data_ptr
    cdef int nb_row , nb_col , nb_disp

    cdef set_data(self, int nb_row, int nb_col , int nb_disp, void * data_ptr):
        self.data_ptr = data_ptr
        self.nb_row = nb_row
        self.nb_col = nb_col
        self.nb_disp = nb_disp


    def __array__(self):
        """
        __Array__ methode called when numpy tries to
            get an array from the object
        """

        cdef np.npy_intp shape[3]
        shape[0] = <np.npy_intp> self.nb_row
        shape[1] = <np.npy_intp> self.nb_col
        shape[2] = <np.npy_intp> self.nb_disp

        #Create a 3D array pf length (size_row,size_col,size_depth)
        ndarray = np.PyArray_SimpleNewFromData(3,shape,np.NPY_FLOAT32, self.data_ptr)
        return ndarray


    def __dealloc__(self):
        """
        Free the array. This is called by Python when all the references
        to the object are gone
        """

        free(<void*>self.data_ptr)


cdef class Array3DWrapperINT:


    cdef void* data_ptr
    cdef int nb_row , nb_col , nb_disp

    cdef set_data(self, int nb_row, int nb_col , int nb_disp, void * data_ptr):
        self.data_ptr = data_ptr
        self.nb_row = nb_row
        self.nb_col = nb_col
        self.nb_disp = nb_disp


    def __array__(self):
        """
        __Array__ methode called when numpy tries to
        get an array from the object
        """

        cdef np.npy_intp shape[3]
        shape[0] = <np.npy_intp> self.nb_row
        shape[1] = <np.npy_intp> self.nb_col
        shape[2] = <np.npy_intp> self.nb_disp

        #Create a 3D array pf length (size_row,size_col,size_depth)
        ndarray = np.PyArray_SimpleNewFromData(3,shape,np.NPY_INT32, self.data_ptr)
        return ndarray


    def __dealloc__(self):
        """
        Free the array. This is called by Python when all the references
        to the object are gone
        """

        free(<void*>self.data_ptr)

# Prototype of C function
cdef extern from "lib/sgm.hpp":

    ctypedef struct CostVolumes:
        float * cost_volume
        int * cost_volume_min



    CostVolumes sgm[T](T * cv_in, T * p1_in,T * p2_in , int* directions, int nb_row, int nb_col, int nb_disp, \
        T invalid_value, bool cost_paths, bool overcounting)

#Create this function is necessary. The goal is to make fused type and memory view work together
#Avoid error as " Cannot coerce to a type that is not specialized"
def sgm_call(my_cv_type[:,:,::1] cv_in_memview, my_cv_type[:,:,::1] p1_in_memview, my_cv_type[:,:,::1] p2_in_memview, \
         int[:,::1] directions, my_cv_type invalid_value, bool cost_paths, bool overcounting):
    """
    Internal function. Wrap the call of sgm C function.

    :param cv_in_memview: cost volume memory view
    :type cv:  3D array contiguous memoryview
    :param p1_in_memview: p1 penalties
    :type p1_in_memview:  3D array contiguous memoryview
    :param p2_in_memview: p2 penalties
    :type p2_in_memview:  3D array contiguous memoryview
    :param directions: directions used in SGM
    :type directions: 2D array
    :param invalid value: value representing invalid cost
    :type invalid_value: uint8 or float
    :param cost_paths: True if Cost Volumes along direction are to be returned
    :type cost_paths: bool
    :param overcounting: over-counting correction option
    :type overcounting: bool
    :return: the aggregated cost volume and the minimum cost along 8 directions
    :rtype: dict of 3D arrays
    """

    nb_row, nb_col, nb_disp = cv_in_memview.shape[0], cv_in_memview.shape[1], cv_in_memview.shape[2]
    nb_dir = 8

    #Float pointer for
    cdef CostVolumes cv_out

    #Calling C function
    cv_out = sgm(&cv_in_memview[0,0,0], &p1_in_memview[0,0,0], &p2_in_memview[0,0,0], &directions[0,0], nb_row, \
                                    nb_col,  nb_disp, invalid_value, cost_paths, overcounting)

    # Build Python object which is an numpy array, data coming from the Cn pointer
    cdef np.ndarray ndarray_out_cv
    array_wrapper_cv = Array3DWrapper()
    array_wrapper_cv.set_data(nb_row, nb_col, nb_disp, < void * > cv_out.cost_volume)
    ndarray_out_cv = np.array(array_wrapper_cv, copy=False)
    # Assign our object to the base of the ndarray object
    ndarray_out_cv.base = < PyObject * > array_wrapper_cv
    # Increment the reference count, as the above assignement was done in C
    # and Python does not know that there is this additional reference
    Py_INCREF(array_wrapper_cv)


    # Build Python object which is an numpy array, data coming from the Cn pointer
    # following variables are defined outside 'if' because of cython constraints.
    cdef np.ndarray ndarray_out_cv_min
    array_wrapper_cv_min = Array3DWrapperINT()
    if (cost_paths):
        array_wrapper_cv_min.set_data(nb_row, nb_col, nb_dir, < void * > cv_out.cost_volume_min)
        ndarray_out_cv_min = np.array(array_wrapper_cv_min, copy=False)
        # Assign our object to the base of the ndarray object
        ndarray_out_cv_min.base = < PyObject * > array_wrapper_cv_min
        # Increment the reference count, as the above assignement was done in C
        # and Python does not know that there is this additional reference
        Py_INCREF(array_wrapper_cv_min)
        cv = {'cv' : ndarray_out_cv, 'cv_min' : ndarray_out_cv_min}
    else :
        cv = {'cv' : ndarray_out_cv, 'cv_min' : None}

    return cv


def sgm_api( cv_in not None, p1_in not None, p2_in not None, direction not None, invalid_value, cost_paths = False, \
                                                                       overcounting = False):
    """
    Compute aggregated cost volume using C++ library where sgm method is implemented

    :param cv: cost volume without aggregation
    :type cv:  3D array contiguous
    :param p1_in: p1 penalties
    :type p1_in:  3D array contiguous
    :param p2_in: p2 penalties
    :type p2_in:  3D array contiguous
    :param directions: directions used in SGM
    :type directions: 2D array
    :param invalid value: value representing invalid cost
    :type invalid value: uint8 or float
    :param cost_paths: True if Cost Volumes along direction are to be returned
    :type cost_paths: bool
    :param overcounting: over-counting correction option
    :type overcounting: bool
    :return: the aggregated cost volume and the minimum cost along 8 directions
    :rtype: dict of 3D arrays
    """

    #Calling wrapper
    ndarray_out = sgm_call(cv_in, p1_in, p2_in, direction,  invalid_value, cost_paths, overcounting)

    return ndarray_out



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of LIBSGM
#
#     https://github.com/CNES/Pandora_libsgm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
import libsgm_python.lr_manager as lrm
import numpy as np


class TestSgmPythonLrManager(unittest.TestCase):

    """"
    Test LrManager used in Python version of LibSGM
    """

    ###############################################################
    # LrManager
    ###############################################################

    def test_dir_1_0_lr_manager(self):
        # Test direction (1, 0)

        shape1 = (4, 3, 3)
        direction1 = [1, 0]
        lr_manager_1 = lrm.LrManager(shape1, direction1)

        seen_pixels = np.zeros((shape1[0], shape1[1]))

        # test init
        self.assertEqual(len(lr_manager_1.planes_front), 1)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [0])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [0, 1, 2])

        current_lr1 = [np.array([[1, 2, 3],
                                [4, 5, 6],
                                [7, 8, 9]])]
        lr_manager_1.set_current_lr(current_lr1)

        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1

        # test next() MOVE Plan
        lr_manager_1.next()
        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1

        self.assertEqual(len(lr_manager_1.planes_front), 1)
        self.assertEqual(len(lr_manager_1.planes_previous), 1)
        # current
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [1])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [0, 1, 2])
        # previous
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["i"], [0])
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["j"], [0, 1, 2])

        # Test lr saving + elimination
        previous_lr1 = np.array([[1, 2, 3],
                                [4, 5, 6],
                                [7, 8, 9]])
        np.testing.assert_array_equal(lr_manager_1.get_previous_lr(0), previous_lr1)

        # test next() MOVE Plan out of dimensions AND add to seen
        lr_manager_1.next()
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        lr_manager_1.next()
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        lr_manager_1.next()

        # out of plan
        self.assertEqual(len(lr_manager_1.planes_front), 0)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)

        # Test if all pixels were seen only once
        np.testing.assert_array_equal(seen_pixels, np.ones((shape1[0], shape1[1])))

    def test_dir_minus1_0_lr_manager(self):
        # Test direction (-1, 0)

        shape1 = (4, 3, 3)
        direction1 = [-1, 0]
        lr_manager_1 = lrm.LrManager(shape1, direction1)

        seen_pixels = np.zeros((shape1[0], shape1[1]))

        # test init
        self.assertEqual(len(lr_manager_1.planes_front), 1)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [3])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [0, 1, 2])

        current_lr1 = [np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9]])]
        lr_manager_1.set_current_lr(current_lr1)

        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1

        # test next() MOVE Plan
        lr_manager_1.next()
        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1

        self.assertEqual(len(lr_manager_1.planes_front), 1)
        self.assertEqual(len(lr_manager_1.planes_previous), 1)
        # current
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [2])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [0, 1, 2])
        # previous
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["i"], [3])
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["j"], [0, 1, 2])

        # Test lr saving + elimination
        previous_lr1 = np.array([[1, 2, 3],
                                [4, 5, 6],
                                [7, 8, 9]])
        np.testing.assert_array_equal(lr_manager_1.get_previous_lr(0), previous_lr1)

        # test next() MOVE Plan out of dimensions AND add to seen
        lr_manager_1.next()
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        lr_manager_1.next()
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        lr_manager_1.next()

        # out of plan
        self.assertEqual(len(lr_manager_1.planes_front), 0)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)

        # Test if all pixels were seen only once
        np.testing.assert_array_equal(seen_pixels, np.ones((shape1[0], shape1[1])))

    def test_dir_0_1_lr_manager(self):
        # Test direction (0,1)

        shape1 = (4, 3, 3)
        direction1 = [0, 1]
        lr_manager_1 = lrm.LrManager(shape1, direction1)

        seen_pixels = np.zeros((shape1[0], shape1[1]))

        # test init
        self.assertEqual(len(lr_manager_1.planes_front), 1)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [0, 1, 2, 3])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [0])

        current_lr1 = [np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9],
                                [10, 11, 12]])]
        lr_manager_1.set_current_lr(current_lr1)

        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1

        # test next() MOVE Plan
        lr_manager_1.next()
        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1

        self.assertEqual(len(lr_manager_1.planes_front), 1)
        self.assertEqual(len(lr_manager_1.planes_previous), 1)
        # current
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [0, 1, 2, 3])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [1])
        # previous
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["i"], [0, 1, 2, 3])
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["j"], [0])

        # Test lr saving + elimination
        previous_lr1 = np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9],
                                [10, 11, 12]])
        np.testing.assert_array_equal(lr_manager_1.get_previous_lr(0), previous_lr1)

        # test next() MOVE Plan out of dimensions AND add to seen
        lr_manager_1.next()
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        lr_manager_1.next()

        # out of plan
        self.assertEqual(len(lr_manager_1.planes_front), 0)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)

        # Test if all pixels were seen only once
        np.testing.assert_array_equal(seen_pixels, np.ones((shape1[0], shape1[1])))

    def test_dir_0_minus1_lr_manager(self):
        # Test direction (0, -1)

        shape1 = (4, 3, 3)
        direction1 = [0, -1]
        lr_manager_1 = lrm.LrManager(shape1, direction1)

        seen_pixels = np.zeros((shape1[0], shape1[1]))

        # test init
        self.assertEqual(len(lr_manager_1.planes_front), 1)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [0, 1, 2, 3])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [2])

        current_lr1 = [np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9],
                                 [10, 11, 12]])]
        lr_manager_1.set_current_lr(current_lr1)

        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1

        # test next() MOVE Plan
        lr_manager_1.next()
        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1

        self.assertEqual(len(lr_manager_1.planes_front), 1)
        self.assertEqual(len(lr_manager_1.planes_previous), 1)
        # current
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [0, 1, 2, 3])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [1])
        # previous
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["i"], [0, 1, 2, 3])
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["j"], [2])

        # Test lr saving + elimination
        previous_lr1 = np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9],
                                 [10, 11, 12]])
        np.testing.assert_array_equal(lr_manager_1.get_previous_lr(0), previous_lr1)

        # test next() MOVE Plan out of dimensions AND add to seen
        lr_manager_1.next()
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        lr_manager_1.next()

        # out of plan
        self.assertEqual(len(lr_manager_1.planes_front), 0)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)

        # Test if all pixels were seen only once
        np.testing.assert_array_equal(seen_pixels, np.ones((shape1[0], shape1[1])))

    def test_dir_1_1_lr_manager(self):
        # Test direction (1, 1)

        shape1 = (4, 3, 3)
        direction1 = [1, 1]
        lr_manager_1 = lrm.LrManager(shape1, direction1)

        seen_pixels = np.zeros((shape1[0], shape1[1]))

        # test init
        self.assertEqual(len(lr_manager_1.planes_front), 2)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)
        # plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [0])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [0, 1, 2])
        # plan 2
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["i"], [1, 2, 3]) # 0 is in plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["j"], [0])

        current_lr1 = [np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9]]),
                       np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9]]) + 12
                       ]
        lr_manager_1.set_current_lr(current_lr1)

        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        seen_pixels[lr_manager_1.planes_front[1]["i"], lr_manager_1.planes_front[1]["j"]] += 1

        # test next() MOVE Plan
        lr_manager_1.next()
        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        seen_pixels[lr_manager_1.planes_front[1]["i"], lr_manager_1.planes_front[1]["j"]] += 1

        self.assertEqual(len(lr_manager_1.planes_front), 2)
        self.assertEqual(len(lr_manager_1.planes_previous), 2)
        # current
        # plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [1])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [1, 2])
        # plan 2
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["i"], [2, 3])
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["j"], [1])
        # previous
        # plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["i"], [0])
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["j"], [0, 1])
        # plan 2
        np.testing.assert_array_equal(lr_manager_1.planes_previous[1]["i"], [1, 2])
        np.testing.assert_array_equal(lr_manager_1.planes_previous[1]["j"], [0])

        # Test lr saving + elimination
        previous_lr1 = np.array([[1, 2, 3],
                                 [4, 5, 6]])
        previous_lr2 = np.array([[1, 2, 3],
                                 [4, 5, 6]]) + 12

        np.testing.assert_array_equal(lr_manager_1.get_previous_lr(0), previous_lr1)
        np.testing.assert_array_equal(lr_manager_1.get_previous_lr(1), previous_lr2)

        # test next() MOVE Plan out of dimensions AND add to seen
        lr_manager_1.next()
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        seen_pixels[lr_manager_1.planes_front[1]["i"], lr_manager_1.planes_front[1]["j"]] += 1
        lr_manager_1.next()

        # out of plan
        self.assertEqual(len(lr_manager_1.planes_front), 0)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)

        # Test if all pixels were seen only once
        np.testing.assert_array_equal(seen_pixels, np.ones((shape1[0], shape1[1])))

    def test_dir_minus1_1_lr_manager(self):
        # Test direction (-1,1)

        shape1 = (4, 3, 3)
        direction1 = [-1, 1]
        lr_manager_1 = lrm.LrManager(shape1, direction1)

        seen_pixels = np.zeros((shape1[0], shape1[1]))

        # test init
        self.assertEqual(len(lr_manager_1.planes_front), 2)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)
        # plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [3])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [0, 1, 2])
        # plan 2
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["i"], [0, 1, 2]) # 3 is in plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["j"], [0])

        current_lr1 = [np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9]]),
                       np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9]]) + 12
                       ]
        lr_manager_1.set_current_lr(current_lr1)

        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        seen_pixels[lr_manager_1.planes_front[1]["i"], lr_manager_1.planes_front[1]["j"]] += 1

        # test next() MOVE Plan
        lr_manager_1.next()
        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        seen_pixels[lr_manager_1.planes_front[1]["i"], lr_manager_1.planes_front[1]["j"]] += 1

        self.assertEqual(len(lr_manager_1.planes_front), 2)
        self.assertEqual(len(lr_manager_1.planes_previous), 2)
        # current
        # plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [2])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [1, 2])
        # plan 2
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["i"], [0, 1])
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["j"], [1])
        # previous
        # plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["i"], [3])
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["j"], [0, 1])
        # plan 2
        np.testing.assert_array_equal(lr_manager_1.planes_previous[1]["i"], [1, 2])
        np.testing.assert_array_equal(lr_manager_1.planes_previous[1]["j"], [0])

        # Test lr saving + elimination
        previous_lr1 = np.array([[1, 2, 3],
                                 [4, 5, 6]])
        previous_lr2 = np.array([[4, 5, 6],
                                 [7, 8, 9]]) + 12

        np.testing.assert_array_equal(lr_manager_1.get_previous_lr(0), previous_lr1)
        np.testing.assert_array_equal(lr_manager_1.get_previous_lr(1), previous_lr2)

        # test next() MOVE Plan out of dimensions AND add to seen
        lr_manager_1.next()
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        seen_pixels[lr_manager_1.planes_front[1]["i"], lr_manager_1.planes_front[1]["j"]] += 1
        lr_manager_1.next()

        # out of plan
        self.assertEqual(len(lr_manager_1.planes_front), 0)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)

        # Test if all pixels were seen only once
        np.testing.assert_array_equal(seen_pixels, np.ones((shape1[0], shape1[1])))

    def test_dir_1_minus1_lr_manager(self):
        # Test direction (1, 1)

        shape1 = (4, 3, 3)
        direction1 = [1, -1]
        lr_manager_1 = lrm.LrManager(shape1, direction1)

        seen_pixels = np.zeros((shape1[0], shape1[1]))

        # test init
        self.assertEqual(len(lr_manager_1.planes_front), 2)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)
        # plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [0])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [0, 1, 2])
        # plan 2
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["i"], [1, 2, 3]) # 0 is in plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["j"], [2])

        current_lr1 = [np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9]]),
                       np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9]]) + 12
                       ]
        lr_manager_1.set_current_lr(current_lr1)

        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        seen_pixels[lr_manager_1.planes_front[1]["i"], lr_manager_1.planes_front[1]["j"]] += 1

        # test next() MOVE Plan
        lr_manager_1.next()
        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        seen_pixels[lr_manager_1.planes_front[1]["i"], lr_manager_1.planes_front[1]["j"]] += 1

        self.assertEqual(len(lr_manager_1.planes_front), 2)
        self.assertEqual(len(lr_manager_1.planes_previous), 2)
        # current
        # plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [1])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [0, 1])
        # plan 2
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["i"], [2, 3])
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["j"], [1])
        # previous
        # plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["i"], [0])
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["j"], [1, 2])
        # plan 2
        np.testing.assert_array_equal(lr_manager_1.planes_previous[1]["i"], [1, 2])
        np.testing.assert_array_equal(lr_manager_1.planes_previous[1]["j"], [2])

        # Test lr saving + elimination
        previous_lr1 = np.array([[4, 5, 6],
                                 [7, 8, 9]])
        previous_lr2 = np.array([[1, 2, 3],
                                 [4, 5, 6]]) + 12

        np.testing.assert_array_equal(lr_manager_1.get_previous_lr(0), previous_lr1)
        np.testing.assert_array_equal(lr_manager_1.get_previous_lr(1), previous_lr2)

        # test next() MOVE Plan out of dimensions AND add to seen
        lr_manager_1.next()
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        seen_pixels[lr_manager_1.planes_front[1]["i"], lr_manager_1.planes_front[1]["j"]] += 1
        lr_manager_1.next()

        # out of plan
        self.assertEqual(len(lr_manager_1.planes_front), 0)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)

        # Test if all pixels were seen only once
        np.testing.assert_array_equal(seen_pixels, np.ones((shape1[0], shape1[1])))

    def test_dir_minus1_minus1_lr_manager(self):
        # Test direction (1, 1)

        shape1 = (4, 3, 3)
        direction1 = [-1, -1]
        lr_manager_1 = lrm.LrManager(shape1, direction1)

        seen_pixels = np.zeros((shape1[0], shape1[1]))

        # test init
        self.assertEqual(len(lr_manager_1.planes_front), 2)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)
        # plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [3])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [0, 1, 2])
        # plan 2
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["i"], [0, 1, 2]) # 3 is in plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["j"], [2])

        current_lr1 = [np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9]]),
                       np.array([[1, 2, 3],
                                 [4, 5, 6],
                                 [7, 8, 9]]) + 12
                       ]
        lr_manager_1.set_current_lr(current_lr1)

        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        seen_pixels[lr_manager_1.planes_front[1]["i"], lr_manager_1.planes_front[1]["j"]] += 1

        # test next() MOVE Plan
        lr_manager_1.next()
        # add to seen
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        seen_pixels[lr_manager_1.planes_front[1]["i"], lr_manager_1.planes_front[1]["j"]] += 1

        self.assertEqual(len(lr_manager_1.planes_front), 2)
        self.assertEqual(len(lr_manager_1.planes_previous), 2)
        # current
        # plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["i"], [2])
        np.testing.assert_array_equal(lr_manager_1.planes_front[0]["j"], [0, 1])
        # plan 2
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["i"], [0, 1])
        np.testing.assert_array_equal(lr_manager_1.planes_front[1]["j"], [1])
        # previous
        # plan 1
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["i"], [3])
        np.testing.assert_array_equal(lr_manager_1.planes_previous[0]["j"], [1, 2])
        # plan 2
        np.testing.assert_array_equal(lr_manager_1.planes_previous[1]["i"], [1, 2])
        np.testing.assert_array_equal(lr_manager_1.planes_previous[1]["j"], [2])

        # Test lr saving + elimination
        previous_lr1 = np.array([[4, 5, 6],
                                 [7, 8, 9]])
        previous_lr2 = np.array([[4, 5, 6],
                                 [7, 8, 9]]) + 12

        np.testing.assert_array_equal(lr_manager_1.get_previous_lr(0), previous_lr1)
        np.testing.assert_array_equal(lr_manager_1.get_previous_lr(1), previous_lr2)

        # test next() MOVE Plan out of dimensions AND add to seen
        lr_manager_1.next()
        seen_pixels[lr_manager_1.planes_front[0]["i"], lr_manager_1.planes_front[0]["j"]] += 1
        seen_pixels[lr_manager_1.planes_front[1]["i"], lr_manager_1.planes_front[1]["j"]] += 1
        lr_manager_1.next()

        # out of plan
        self.assertEqual(len(lr_manager_1.planes_front), 0)
        self.assertEqual(len(lr_manager_1.planes_previous), 0)

        # Test if all pixels were seen only once
        np.testing.assert_array_equal(seen_pixels, np.ones((shape1[0], shape1[1])))



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of LIBSGM
#
#     https://github.com/CNES/Pandora_libsgm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions to execute python SGM library.
"""

import numpy as np
import warnings

from .lr_manager import LrManager


def run_sgm(cv_in: np.ndarray, p1_in: np.ndarray, p2_in: np.ndarray, directions: np.ndarray, cost_paths: bool = False,
            overcounting: bool = False):
    """
    Run Python LibSGM

    :param cv_in: cost volume
    :type cv_in:  np.ndarray
    :param p1_in: p1 penalties
    :type p1_in:  3D np.ndarray
    :param p2_in: p2 penalties
    :type p2_in:  3D np.ndarray
    :param directions: directions used in SGM
    :type directions: 2D np.ndarray
    :param cost_paths: True if Cost Volumes along direction are to be returned
    :type cost_paths: bool
    :param overcounting: over-counting correction option
    :type overcounting: bool
    :return: the aggregated cost volume and the minimum cost along directions
    :rtype: dict of 3D np.ndarray
    """

    cv_out = {"cv": np.zeros(cv_in.shape)}

    if cost_paths:
        cv_out["cv_min"] = np.zeros((cv_in.shape[0], cv_in.shape[1], len(directions)))

    for i in range(len(directions)):
        # optimize this direction
        lr_manager = LrManager(cv_in.shape, directions[i])

        nb_planes = len(lr_manager.planes_front)
        while nb_planes > 0:
            current_lr = []

            for p in range(nb_planes):
                front_plane = lr_manager.planes_front[p]
                if len(lr_manager.planes_previous) == 0:
                    # initialized, previous plane doesnt exist
                    partial_lr = cv_in[front_plane["i"], front_plane["j"], :]
                    cv_out["cv"][front_plane["i"], front_plane["j"], :] += partial_lr
                    current_lr.append(partial_lr)
                else:
                    partial_lr = np.zeros(cv_in[front_plane["i"], front_plane["j"], :].shape)
                    previous_lr = lr_manager.get_previous_lr(p)
                    for d in range(cv_in.shape[2]):
                        partial_lr[:, d] = compute_lr(
                            cv_in[front_plane["i"], front_plane["j"], :], previous_lr, d,
                            p1_in[front_plane["i"], front_plane["j"], i], p2_in[front_plane["i"], front_plane["j"], i])

                    cv_out["cv"][front_plane["i"], front_plane["j"]] += partial_lr
                    current_lr.append(partial_lr)

                if cost_paths:
                    cv_out["cv_min"][front_plane["i"], front_plane["j"], i] = np.argmin(partial_lr, axis=1)

            lr_manager.set_current_lr(current_lr)

            # compute next planes
            lr_manager.next()
            nb_planes = len(lr_manager.planes_front)

    if overcounting:
        cv_out["cv"] -= (len(directions) - 1) * cv_in

    return cv_out


def compute_lr(cv_in_2d_front: np.ndarray, lr_2d_previous: np.ndarray, d: int, p1_in_1d: np.ndarray,
               p2_in_1d: np.ndarray):
    """
    Compute Lr of current plane, at a given disparity

    :param cv_in_2d_front: cost volume
    :type cv_in_2d_front:  np.ndarray
    :param lr_2d_previous: previous lr computed
    :type lr_2d_previous: np.ndarray
    :param d: current disparity
    :type d:  int
    :param p1_in_1d: p1 penalties, dim=1
    :type p1_in_1d: np.ndarray
    :param p2_in_1d: p2 penalties, dim=1
    :type p2_in_1d: np.ndarray
    :return: partial lrs
    :rtype: np.ndarray
    """

    # Assign penalties
    penalties = np.repeat(p2_in_1d[:, np.newaxis], cv_in_2d_front.shape[1], axis=1)
    penalties[:, d] = 0
    if (d - 1) >= 0:
        penalties[:, d - 1] = p1_in_1d
    if (d + 1) < cv_in_2d_front.shape[1]:
        penalties[:, d + 1] = p1_in_1d

    min_previous_lr_penalties = np.nanmin(lr_2d_previous + penalties, axis=1)
    min_previous_lr = np.nanmin(lr_2d_previous, axis=1)
    indexes_nan = np.where(np.isnan(min_previous_lr_penalties))
    min_previous_lr_penalties[indexes_nan] = 0
    min_previous_lr[indexes_nan] = 0

    return cv_in_2d_front[:, d] + min_previous_lr_penalties - min_previous_lr



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of LIBSGM
#
#     https://github.com/CNES/Pandora_libsgm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains the required libraries and softwares allowing to execute the software, and setup elements to configure and identify the software. 
"""

from setuptools import setup, find_packages
import os
import shutil
from codecs import open
import sys

try:
    import numpy
except ImportError:
    print("")
    print("WARNING ! Installation of numpy is required before libSGM installation")
    print("")
    raise

try:
    from Cython.Distutils.extension import Extension
except ImportError:
    from setuptools import Extension
    from setuptools.command.build_ext import build_ext
    USING_CYTHON = False
else:
    from Cython.Distutils import build_ext
    USING_CYTHON = True


cmdclass = {'build_ext': build_ext}

def readme():
    with open('README.md', "r", "utf-8") as f:
        return f.read()


src_dir = "sources"

if USING_CYTHON:
    sources = [src_dir + "/lib/sgm.cpp", src_dir+"/sgm_wrapper.pyx"]
else:
    sources = [src_dir + "/lib/sgm.cpp", src_dir+"/sgm_wrapper.cpp"]


ext_1 = Extension('libSGM.sgm_wrapper',
                  sources,
                  language='c++',
                  library_dirs=[],
                  libraries=[],
                  include_dirs=[numpy.get_include(), src_dir + "/lib/sgm.hpp" ],
                  extra_compile_args=['-O3', '-fopenmp' ,'-std=c++11'],
                  extra_link_args=['-lgomp']
                  )

extensions = [ext_1]



try:
    from sphinx.setup_command import BuildDoc
    cmdclass.update({'build_sphinx': BuildDoc})
except ImportError:
    print('WARNING: sphinx not available. Doc cannot be built')


os.environ['CC'] = shutil.which('gcc')
os.environ['CXX'] = shutil.which('g++')

requirements = ['numpy',
                'nose2']

setup(name='libSGM',
      version='0.3.1',
      description='libSGM is a CNES version of H.Hirschmuller Semi-Global Matching',
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='https://github.com/CNES/Pandora_libsgm',
      author='CNES',
      author_email='myriam.cournet@cnes.fr',
      license='Apache License 2.0',
      zip_safe=False,
      packages=find_packages(),
      ext_modules=extensions,
      cmdclass=cmdclass,
      command_options={
          'build_sphinx': {
              'build_dir': ('setup.py', 'doc/build/'),
              'source_dir': ('setup.py', 'doc/source/')}},
      entry_points={
              'libsgm': [
              'python_libsgm = libsgm_python.sgm_python:run_sgm'
              ],
      },
      install_requires=requirements
      )



#!/usr/bin/env python
# coding: utf8
#
# Copyright (c) 2020 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of LIBSGM
#
#     https://github.com/CNES/Pandora_libsgm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains functions to execute python SGM library.
"""

import numpy as np
from numba import njit, prange
from typing import List

from .lr_manager import LrManager


def run_sgm_parall(cv_in: np.ndarray, p1_in: np.ndarray, p2_in: np.ndarray, directions: np.ndarray,
                   cost_paths: bool = False, overcounting: bool = False):
    """
    Run Python LibSGM

    :param cv_in: cost volume
    :type cv_in:  np.ndarray
    :param p1_in: p1 penalties
    :type p1_in:  3D np.ndarray
    :param p2_in: p2 penalties
    :type p2_in:  3D np.ndarray
    :param directions: directions used in SGM
    :type directions: 2D np.ndarray
    :param cost_paths: True if Cost Volumes along direction are to be returned
    :type cost_paths: bool
    :param overcounting: over-counting correction option
    :type overcounting: bool
    :return: the aggregated cost volume and the minimum cost along 8 directions
    :rtype: dict of 3D np.ndarray
    """

    # Compute initial points
    # [i0, j0, di, dj, num_dir]
    starting_points = []
    for i in range(len(directions)):
        d = directions[i]
        # optimize this direction
        lr_manager = LrManager(cv_in.shape, directions[i])

        for plane in lr_manager.planes_front:
            list_indexes = [[i0, j0, d[0], d[1], i] for idx1, i0 in enumerate(plane["i"]) for idx2, j0 in
                            enumerate(plane["j"])]
            starting_points += list_indexes

    # compute costs for all
    dir_str = []
    for i in range(len(directions)):
        dir_str.append(str(i))
    dir_str = np.array(dir_str)
    cv_out = compute_costs(np.array(starting_points), cv_in, p1_in, p2_in, dir_str, cost_paths=cost_paths)

    if overcounting:
        cv_out["cv"] -= (len(directions) - 1) * cv_in

    return cv_out


@njit(parallel=True, cache=False)
def compute_costs(starting_points: List, cv_in: np.ndarray, p1_in: np.ndarray, p2_in: np.ndarray, dir_str: np.ndarray,
                  cost_paths: bool = False):
    """
    Compute cost volume, starting from given points / directions

    :param starting_points: List of starting points [i0, j0, dir_i, dir_j, num_dir]
    :type starting_points:  List
    :param cv_in: cost volume
    :type cv_in:  3D np.ndarray
    :param p1_in: p1 penalties
    :type p1_in:  3D np.ndarray
    :param p2_in: p2 penalties
    :type p2_in:  3D np.ndarray
    :param dir_str: list of direction names
    :type dir_str:  List
    :param cost_paths: True if Cost Volumes along direction are to be returned
    :type cost_paths: bool
    :return: the aggregated cost volume and the minimum cost along directions
    :rtype: dict of 3D np.ndarray
    """
    in_shape = cv_in.shape
    cv = {"cv": np.zeros(in_shape)}

    for i in range(len(dir_str)):
        name = "cv_" + dir_str[i]
        cv[name] = np.zeros(in_shape)

    cv_min = np.zeros((in_shape[0], in_shape[1], p1_in.shape[2]))

    for p in prange(starting_points.shape[0]):
        point = starting_points[p]

        d_i = point[2]
        d_j = point[3]
        i_0 = point[0]
        j_0 = point[1]
        dir = point[4]
        name = "cv_" + dir_str[dir]

        cv["cv"][i_0, j_0, :] += cv_in[i_0, j_0, :].copy()
        cv[name][i_0, j_0, :] += cv_in[i_0, j_0, :].copy()
        previous_lr = cv_in[i_0, j_0, :].copy()

        i = i_0 + d_i
        j = j_0 + d_j

        while 0 <= i < in_shape[0] and 0 <= j < in_shape[1]:

            current_lr = cv_in[i, j, :].copy()

            p1 = p1_in[i, j, dir]
            p2 = p2_in[i, j, dir]

            for d in range(in_shape[2]):
                penalties = p2 * np.ones(in_shape[2])
                penalties[max(0, d-1): min(d+1, in_shape[2]-1) + 1] = p1
                penalties[d] = 0

                val = np.nanmin(penalties + previous_lr) - np.nanmin(previous_lr)
                if not np.isnan(val):
                    current_lr[d] += val

            # save current lr
            cv["cv"][i, j, :] += current_lr
            cv[name][i, j, :] += current_lr

            # next pixel
            previous_lr = current_lr

            cv_min[i, j, dir] = np.argmin(current_lr)

            i += d_i
            j += d_j

    if cost_paths:
        cv["cv_min"] = cv_min

    return cv
